(* type width *)
type width = int
(* type height *)
type height = int
(* type t *)
type t = (width * height * Strng.t list)
(* height : (width * height * Strng.t list) -> height *)
let e = (0,0,[[]])
let height (_,h,_) = h
(* width : (width * height * Strng.t list) -> width *)
let width (w,_,_) = w
(* above : t -> t -> t *)
let above (w1,h1,p1) (w2,h2,p2) =
 if w1 = w2 then (w1,h1+h2,p1@p2)
            else failwith "different widths"
(* beside : t -> t -> t *)
let beside (w1,h1,p1) (w2,h2,p2) =
 if h1 = h2 then (w1+w2,h1,Lst.zip_with (@) p1 p2)
            else failwith "different heights"
(* to_strng : t -> Strng.t *)
let to_strng (_,_,p) = Lst.join ['\n'] p 
(* stack : t list -> t *)
let stack ps = Lst.foldr1 above ps
(* spread : t list -> t *)
let spread ps = Lst.foldr1 beside ps
(* pixel : char -> t *)
let pixel c = (1,1,[[c]])
(* row : char list -> t *)
let row s = spread(Lst.map pixel s)
(* column : char list -> t *)
let column s = stack(Lst.map pixel s)
(* empty : width -> height -> t *)
let empty w h =
 let line = Lst.replicate w ' ' in
 let rows = Lst.replicate h line in
 stack(Lst.map row rows)
(* stack_with : height -> t list -> t *)
let stack_with h ps = Lst.foldr1 (fun p q ->
 above (above p (empty (width q) h)) q) ps
(* spread_with : width -> t list -> t *)
let spread_with w ps = Lst.foldr1 (fun p q ->
 beside (beside p (empty w (height q))) q) ps
(* tile : t list list -> t *)
let tile pss = stack(Lst.map spread pss)
(* tile_with : width -> height -> t list list -> t *)
let tile_with w h pss = 
 stack_with h (Lst.map (spread_with w) pss)
(* toplevel_printer : Format.formatter -> t -> unit *)
let toplevel_printer fmt ((_,_,lines):t) =
 Format.fprintf fmt "\n";
 List.iter (fun line ->
  List.iter (Format.fprintf fmt "%c") line; Format.fprintf fmt "\n"
 ) lines
